#
# A test code of EM algorithm for mixture models
# inspired by Shotaro Akaho's code, but made from scratch
#
# 
emstep<-function(model,...){
  UseMethod('emstep')
}
plot3<-function(model,...){
  UseMethod('plot3')
}
dprob<-function(model,...){
  UseMethod('dprob')
}
rprob<-function(model,...){
  UseMethod('rprob')
}
mle<-function(model,...){
  UseMethod('mle')
}
loglike<-function(model,...){
  UseMethod('loglike')
}

# model class declaration
gauss<-function(d){ # d-dimensional Gaussian
  mu<-rep(0,d)
  sigma<-diag(d)
  model<-list(mu=mu,sigma=sigma)
  class(model)<-'gauss'
  return(model)
}

gaussd<-function(d){ # d-dimensional Gaussian with diagonal variance
  mu<-rep(0,d)
  sigma<-rep(1,d)
  model<-list(mu=mu,sigma=sigma)
  class(model)<-'gaussd'
  return(model)
}

gaussu<-function(d){ # d-dimensional Gaussian with uniform variance
  mu<-rep(0,d)
  sigma<-1
  model<-list(mu=mu,sigma=sigma)
  class(model)<-'gaussu'
  return(model)
}

lreg<-function(d){ # d-dimensional linear regressor
  beta<-rep(0,d+1)
  sigma<-1
  model<-list(beta=beta,sigma=sigma)
  class(model)<-'lreg'
  return(model)
}

gaussreg<-function(gauss,reg){ # Gaussian with regressor for last dim
  model<-list(gauss=gauss,reg=reg)
  class(model)<-'gaussreg'
  return(model)
}

mixture<-function(models,prob=rep(1/length(models),length(models))){ # mixture
  tmp<-list(prob=prob,models=models)
  class(tmp)<-"mixture"
  return(tmp)
}

# 2D plotter
plot.gauss<-function(model,add=F,col="blue"){
  s<-svd(model$sigma)
  q<-s$v%*%diag(sqrt(s$d))%*%t(s$v)
  t<-seq(0,2*pi,length=36) 
  if(add==F) {
    plot(rprob(model,50),type="n",xlab="x",ylab="y")
  }
  lines(t(q%*%t(cbind(cos(t),sin(t)))+model$mu),col=col)
  points(t(model$mu),col=col,pch="*")
}

plot.gaussd<-function(model,add=F,col="blue"){
  d<-length(model$mu)
  tmp<-gauss(d)
  tmp$mu<-model$mu
  tmp$sigma<-diag(model$sigma)
  plot(tmp,add=add,col=col)
}

plot.gaussu<-function(model,add=F,col="blue"){
  d<-length(model$mu)
  tmp<-gauss(d)
  tmp$mu<-model$mu
  tmp$sigma<-model$sigma*diag(d)
  plot(tmp,add=add,col=col)
}

plot.mixture<-function(model,add=F,col="blue"){
  if(add==F) {
    plot(rprob(model,50),type="n",xlab="x",ylab="y")
  }
  lapply(model$models,function(model){plot(model,add=TRUE,col=col)})
}

# 3D plotter
plot3.default<-function(model,xlim=c(-10,10),ylim=c(-10,10)){
  n<-40
  x<-seq(xlim[1],xlim[2],length=n)
  y<-seq(ylim[1],ylim[2],length=n)
  xx<-rep(x,n)
  yy<-rep(y,each=n)
  z<-matrix(dprob(model,cbind(xx,yy)),n,n)
  persp(x,y,z,theta=30,phi=30,expand=0.5,col="lightblue")
}

# probability density
dprob.gauss<-function(model,x){
  x<-as.matrix(x)
  d<-ncol(x)
  n<-nrow(x)
  if(d==1) {
    z<-1/sqrt(2*pi*model$sigma)*exp(-mahalanobis(x,model$mu,model$sigma)/2)
  } else {
    z<-1/sqrt((2*pi)^d*det(model$sigma))*exp(-mahalanobis(x,model$mu,model$sigma)/2)
  }
  return(matrix(z,n,1))
}

dprob.gaussd<-function(model,x){
  d<-length(model$mu)
  tmp<-gauss(d)
  tmp$mu<-model$mu
  tmp$sigma<-diag(model$sigma)
  return(dprob(tmp,x))
}

dprob.gaussu<-function(model,x){
  d<-length(model$mu)
  tmp<-gauss(d)
  tmp$mu<-model$mu
  tmp$sigma<-model$sigma*diag(d)
  return(dprob(tmp,x))
}

dprob.lreg<-function(model,x,y){
  tmp<-gauss(1)
  tmp$sigma<-model$sigma
  xx<-cbind(x,rep(1,nrow(x)))
  beta<-as.matrix(model$beta)
  res<-y-drop(xx%*%beta)
  return(dprob(tmp,res))
}

dprob.gaussreg<-function(model,x){
  d<-ncol(x)
  xx<-x[,1:(d-1)]
  y<-x[,d]
  return(dprob(model$gauss,xx)*dprob(model$reg,xx,y))
}

dprob.mixture<-function(model,x){
  k<-length(model$models)
  x<-as.matrix(x)
  d<-ncol(x)
  n<-nrow(x)
  z<-matrix(unlist(lapply(model$models,function(model){dprob(model,x)})),n,k)
  return(apply(sweep(z,2,model$prob,"*"),1,sum))
}

# random generation
rprob.gauss<-function(model,n){
  d<-length(model$mu)
  s<-svd(model$sigma)
  q<-s$v%*%diag(sqrt(s$d))%*%t(s$v)
  z<-matrix(rnorm(d*n),n,d)
  return(sweep(z%*%q,2,model$mu,"+"))
}

rprob.gaussd<-function(model,n){
  d<-length(model$mu)
  tmp<-gauss(d)
  tmp$mu<-model$mu
  tmp$sigma<-diag(model$sigma)
  return(rprob(tmp,n))
}

rprob.gaussu<-function(model,n){
  d<-length(model$mu)
  tmp<-gauss(d)
  tmp$mu<-model$mu
  tmp$sigma<-model$sigma*diag(d)
  return(rprob(tmp,n))
}

rprob.mixture<-function(model,n){
  k<-length(model$models)
  j<-sample(1:k,n,replace=TRUE,prob=model$prob)
  z<-NULL
  for(i in 1:k){
    m<-sum(j==i)
    if(m>0){
      z<-rbind(z,rprob(model$models[[i]],m))
    }
  }
  return(z)
}

# some functions for EM algorithm
loglike<-function(model,x){
  return(sum(log(dprob(model,x))))
}

cprob<-function(model,x){
  k<-length(model$models)
  x<-as.matrix(x)
  d<-ncol(x)
  n<-nrow(x)
  p<-matrix(unlist(lapply(model$models,function(model){dprob(model,x)})),n,k)
  # get p(x|k)
  p<-sweep(p,2,model$prob,"*") # get p(x,k)
  q<-apply(p,1,sum) # get p(x)
  return(sweep(p,1,q,"/"))
}

emstep.mixture<-function(model,x){
  k<-length(model$models)
  x<-as.matrix(x)
  d<-ncol(x)
  n<-nrow(x)
  p<-cprob(model,x)
  for(i in 1:k) {
    model$models[[i]]<-mle(model$models[[i]],x,weight=drop(p[,i]))
  }
  model$prob<-apply(p,2,sum)/n
  return(model)
}

mle.gauss<-function(model,x,weight=NULL){
  x<-as.matrix(x)
  d<-ncol(x)
  n<-nrow(x)
  if(is.null(weight)){
    mu<-apply(x,2,sum)/n
    sigma<-crossprod(x)/n-mu%o%mu
#    sigma<-t(x)%*%x/n-mu%o%mu
  } else {
    z<-sum(weight)
    mu<-apply(weight*x,2,sum)/z
    sigma<-crossprod(x,weight*x)/z-mu%o%mu
#    sigma<-t(x)%*%(weight*x)/z-mu%o%mu    
  }
  model$mu<-drop(mu)
  model$sigma<-sigma+diag(10^{-8},d)
  return(model)
}

mle.gaussd<-function(model,x,weight=NULL){
  d<-length(model$mu)
  tmp<-gauss(d)
  tmp<-mle(tmp,x,weight=weight)
  model$mu<-tmp$mu
  model$sigma<-diag(tmp$sigma)
  return(model)
}

mle.gaussu<-function(model,x,weight=NULL){
  d<-length(model$mu)
  tmp<-gauss(d)
  tmp<-mle(tmp,x,weight=weight)
  model$mu<-tmp$mu
  model$sigma<-mean(diag(tmp$sigma))
  return(model)
}

mle.lreg<-function(model,x,y,weight=NULL){
  x<-as.matrix(x)
  d<-ncol(x)
  n<-nrow(x)
  xx<-cbind(x,rep(1,n))
  if(is.null(weight)){
    beta<-solve(crossprod(xx))%*%crossprod(xx,y)
    res<-y-xx%*%beta
    sigma<-crossprod(res)/n
  } else {
    z<-sum(weight)
    beta<-solve(crossprod(xx,weight*xx))%*%crossprod(xx,weight*y)
    res<-y-xx%*%beta
    sigma<-crossprod(res,weight*res)/z
  }
  model$beta<-drop(beta)
  model$sigma<-sigma+10^{-8}
  return(model)
}

mle.gaussreg<-function(model,x,y,weight=NULL){
  x<-as.matrix(x)
  d<-ncol(x)
  n<-nrow(x)
  xx<-x[,1:(d-1)]
  y<-x[,d]
  model$gauss<-mle(model$gauss,xx,weight=weight)
  model$reg<-mle(model$reg,xx,y,weight=weight)
  return(model)
}

