adaboost.m2 <- function(x,y,T=100,wl="stump.m",na.action="remove",
                        previous.res=NULL,Y.num=1,r.wl=0.5,r.b=0.1){
  ## アダブースト.M2でGクラス問題に対して判別器を構成する関数
  ## 入力 (x: 入力,y: ラベル, T: ステップ数,wl: 弱学習器, スタンプを使用,
  ##      na.action: xにNAが含まれている場合の対応("remove","all"),
  ##      previous.res[adaboost.m2の結果]: 指定した学習結果から学習を開始,
  ##      Y.num[ベクトル]: スタンプにおいて, 全G個のクラスのうち，
  ##                      Y.num個のクラスと(G-Y.num)個のクラスの2値判別を考える,
  ##      r.wl[0<r.wl<=1の実数]: Y.numで決まる2値判別問題のうち用いる割合,
  ##      r.b[0<r.wl<=1の実数]: スタンプにおいて閾値決定に用いるサンプルの割合)
  ## 出力 (class: 予測されたクラスラベル, error: 誤り率, loss: ロス関数の値,
  ##      F: 判別関数の値, alpha: 弱学習器の係数, WL: 弱学習器, INFO: x,yの情報)
  
  INFO <- make.info(x,y) # データの情報を作成
  dat <- trans.data(x,y,INFO,na.action) # データの加工
  INFO$na.index <- dat$na.index
  INFO$na.action <- na.action
  # adaboost.m2を学習
  res <- learn.adaboost.m2(dat$x,dat$y,T,wl,Y.num=Y.num,
                           previous.res,r.wl=r.wl,r.b=r.b)
  if(na.action=="remove"){ # 欠測を除いて学習した場合は，欠測したデータの予測も行う
    if(length(INFO$na.index)>0){
      F.all <- matrix(0,length(y),length(INFO$level.y))
      dat <- trans.data(x[,,drop=FALSE],y,INFO,"all")
      F.all <- predict(res,dat$x,dat$y,T)$F
      res$F <- F.all
    }
  }
  res$INFO <- INFO
  res$class <- INFO$level.y[max.col(res$F)]
  class(res) <- 'adaboost.m2'
  return(res)
}

learn.adaboost.m2 <- function(x,y,T,wl,Y.num,previous.res,r.wl,r.b){
  ## アダブースト.M2を学習する関数
  ## 入力 (xが数値に変換された行列である以外は, 関数adaboost.m2()の入力と同じ)
  ## 出力 (error: 訓練誤差の系列, loss: ロス関数の系列,　F: データに対する判別関数の値,
  ##      alpha: 弱仮説の係数, WL: 弱仮説)
  t0 <- 1 
  p <- ncol(x)
  n <- nrow(x)
  class <- length(unique(y))
  error <- rep(0,T)
  loss <- rep(0,T)
  alpha <- rep(0,T)
  WL <- list()
  w <- matrix(1,n,class)
  w[cbind(c(1:n),y)] <- 0
  w <- w/sum(w)
  F <- matrix(0,n,class)

  if(!is.null(previous.res)){ # 以前に学習した判別関数から学習をスタートする場合
    T.prev <- length(previous.res$alpha)
    if(T.prev >=T){
      return(previous.res)
    }
    else if(T.prev >length(previous.res$WL)){
      cat("Weighted error of weak hypothesis is 0\n")
      return(previous.res)
    }
    else{
      t0 <- T.prev+1
      error <- previous.res$error
      loss <- previous.res$loss
      WL <- previous.res$WL
      alpha[1:T.prev] <- previous.res$alpha
      for(t in 1:T.prev){
        F <- F+alpha[t]*predict(WL[[t]],x)
      }
      w <- exp(F-F[cbind(c(1:n),y)])
      w[cbind(c(1:n),y)] <- 0
      w <- w/sum(w)  # 重みの正規化
    }
  }

  for(t in t0:T){
    WL[[t]] <- list()
    class(WL[[t]]) <- wl
    WL[[t]] <- learn.stump.m(WL[[t]],x,y,w,Y.num,r.wl,r.b) # 弱仮説の学習
    if(WL[[t]]$E==0){ # 誤り率が0であればアルゴリズムを停止
      alpha[t] <- 0.5*log((1-min(w)/100)/(min(w)/100))
      F <- alpha[t]*predict(WL[[t]],x)
      error[t:T] <-  mean(sign(y != sign(F) ))
      loss[t:T] <- mean( exp(F-F[cbind(c(1:n),y)]) )
      cat(c("\n",paste("Error rate of weak hypothesis is 0 at step",t)))
      break
    }
    else{
      alpha[t] <- 0.5*log((1-WL[[t]]$E)/WL[[t]]$E) # 係数の計算
      F <- F+alpha[t]*predict(WL[[t]],x) # 判別関数の更新
      w <- exp(F-F[cbind(c(1:n),y)]) # 重みの更新
      w[cbind(c(1:n),y)] <- 0
      w <- w/sum(w) # 正規化
      error[t] <-  mean(sign(y != max.col(F) )) # 誤り率の計算
      loss[t] <- mean( exp(F-F[cbind(c(1:n),y)]) ) # ロス関数の計算
    }
  }
  res <-  list(error=error,loss=loss,F=F,alpha=alpha,WL=WL)
  class(res) <- 'adaboost.m2'
  return(res)
}

predict.adaboost.m2 <- function(res,x,y=NULL,T=length(res$WL)){
  ## アダブースト.M2の結果を用いてラベルを予測する関数
  ## 入力 (res: adaboost.m2の結果，x: 入力,y: ラベル,T: ステップ数)
  ## 出力 (F: 判別関数の値, class: 予測されたクラスラベル, error: 誤り率)
  
  F <- matrix(0,nrow(x),length(res$INFO$level.y))
  error <- NULL
  if(is.null(y)){ # 真のラベルが与えられない場合
    dat <- trans.data(x,y=NULL,res$INFO,"all")
  }
  else{ # 真のラベルが与えられている場合
    error <- rep(0,T)
    dat <- trans.data(x,y,res$INFO,"all")
  }
  for(t in 1:T){
    F <- F+res$alpha[t]*predict(res$WL[[t]],dat$x)
    if(!is.null(y)){ # yが与えられている時は誤り率を計算
      error[t] <- mean(max.col(F)!=dat$y)
    }
  }
  class <- res$INFO$level.y[max.col(F)] # ラベル予測
  return(list(F=F,class=class,error=error))
}

learn.stump.m <- function(WL,x,y,w,Y.num,r.wl,r.b){
  ## adaboost.m2で，スタンプを学習するための関数
  ## 入力 (WL: stump.mというクラスのリスト, w: 例題の重み，
  ##      その他はadaboost.m2()の入力に準じる)
  ## 出力 (b: 閾値,S: スタンプの符号,E: 重み付き誤り率,
  ##      idx: 選択された次元,set: クラスセットC, class: クラス数)
  
  n <- nrow(x)
  p <- ncol(x)
  class <- length(unique(y))
  e.min <- 1

  class.set <- function(NUM,l){
    ## クラスセット{1,...,G}からNUM個のクラスを選択するローカル関数
    ## 入力 (NUM: 選択する数, l: 乱数の種)
    ## 出力 (選択されたクラス)
    tmp <- NULL
    for(i in 1:class){
      tmp <- paste(tmp,"c(1,0),",sep="")
    }
    tmp <- substr(tmp,1,nchar(tmp)-1)
    Cand <- eval(parse(text=paste("expand.grid(",tmp,")",sep="")))
    which(Cand[apply(Cand,1,sum)==NUM,][l,]!=0)
  }

  for(k in Y.num){
    N <- prod(c(class:(class-k+1)))/prod(c(1:k))
    N <- sort(sample(N,N*r.wl))
    for(l in N){
      SET <- class.set(k,l)
      ynew <- apply(outer(y,SET,function(x,y){x==y}),1,sum)*2-1
      Y <- rep(-1,class)
      Y[SET] <- 1
      for(j in 1:p){
        xx <- sort(unique(x[!is.na(x[,j]),j]))
        if(length(xx)*r.b >10){
          xx.set <- sort(sample(length(xx),(length(xx)*r.b) %/% 1))
          xx <- xx[xx.set]        
        }
        for(i in 1:(length(xx)-1)){
          sgn <- 1
          delta <- (xx[i]+xx[i+1])/2
          fval <- sgn*sign(x[,j]-delta)
          fval[is.na(fval)] <- 0
          e.current <- sum(w*(1+outer(fval,Y,function(x,y){sign(x==y)})-
                              sign(fval==ynew))/2)
          if(e.current > 0.5){
            e.current <- 1-e.current
            sgn <- -sgn
          }
          if(e.current < e.min){
            b <- delta
            ss <- sgn
            e.min <- e.current
            index <- j
            set.index <- SET
          }
        } # i end
      }
    }
  }
  res <- list(b=b,S=ss,E=e.min,idx=index,set=set.index,class=class)
  class(res) <- class(WL)
  return(res)
}  ## stump end

predict.stump.m <- function(WL,x){
  ## 弱仮説stump.mの予測をする関数
  ## 入力 (WL: learn.stump.m()の結果,x:予測したい入力)
  ## 出力 (xに対する弱仮説の予測値)
  yhat <- WL$S*sign(x[,WL$idx]-WL$b)
  yhat[is.na(yhat)] <- 0
  Y <- rep(-1,WL$class)
  Y[WL$set] <- 1
  outer(yhat,Y,function(x,y){sign(x==y)})
}
