library(kernlab)
mpm.rbf <- function(x, y, sigma=mean(sigest(x,scaled=FALSE)),
                    maxit=100,delta=1/nrow(x),tol=.Machine$double.eps){
  # ガウシアン・カーネルによるミニマックス確率マシンで判別を行う関数
  # 入力(x[行列]:入力データ, y:出力ラベル, 
  #      sigma:ガウシアン・カーネルのパラメータ
  #      maxit:ブロック座標降下法の繰り返し数の最大値，
  #      delta:正則化項のパラメータ，  
  #      tol:ブロック座標降下法の停止基準)
  # 出力(f:線形判別関数, lab:ラベル, worst:最悪の予測誤差の推定値)
  n <- nrow(x)
  x1 <- x[y==levels(y)[1],]; x2 <- x[y==levels(y)[2],]
  K11 <- kernelMatrix(rbfdot(sigma=sigma),x1)
  K12 <- kernelMatrix(rbfdot(sigma=sigma),x1,x2)
  K21 <- t(K12)
  K22 <- kernelMatrix(rbfdot(sigma=sigma),x2)
  m1 <- rowMeans(rbind(K11,K21))
  m2 <- rowMeans(rbind(K12,K22))
  v1 <- (rbind(K11,K21) %*% t(rbind(K11,K21)))/nrow(x1)- m1 %o% m1
  v2 <- (rbind(K12,K22) %*% t(rbind(K12,K22)))/nrow(x2)- m2 %o% m2
  ## 期待，分散共分散行列の計算
  a0 <- (m1-m2)/sum((m1-m2)^2)
  F0 <- diag(n)-(a0 %o% a0)/sum(a0^2)
  eigenF <- eigen(F0)
  F <- matrix(eigenF$vector[,order(eigenF$value)[-1]],n,n-1)
  G <- t(F) %*% v1 %*% F; H <- t(F) %*% v2 %*% F
  g <- t(F) %*% v1 %*% a0; h <- t(F) %*% v2 %*% a0
  ## 初期化
  beta <- 1; eta <- 1; k <- 0
  err <- Inf; obj_pre <- Inf
  ## ブロック座標降下法
  while((err > tol) & (k < maxit)){
    MLS <- G/beta + H/eta + delta*diag(n-1)
    bLS <- -(g/beta+h/eta)
    uLS <- solve(MLS,bLS)
    w <- a0 + F %*% uLS
    beta <- sqrt(sum((v1 %*% w) * w))
    eta <- sqrt(sum((v2 %*% w) * w))
    obj_val <- beta + eta + sum(w^2)*delta/2
    err <- abs((obj_pre - obj_val)/obj_val)
    obj_pre <- obj_val
    k <- k+1
  }
  ## 推定値の計算
  b <- sum(w * m1) - beta/(beta+eta)
  gamma <- beta+eta
  worst_err <- (gamma^2)/(1+gamma^2)
  res <- list(f=function(z)
              kernelMatrix(rbfdot(sigma=sigma),z,rbind(x1,x2))%*%w-b,
              sigma=sigma, lab=levels(y), worst=worst_err)
  class(res) <- 'mpm'
  res
}
