mpm <- function(x,y,maxit=100,delta=10^(-8),
                tol=.Machine$double.eps){
  # ミニマックス確率マシンによる線形判別を行う関数
  # 入力(x[行列]:入力データ, y:出力ラベル,
  #      maxit:ブロック座標降下法の繰り返し数の最大値，
  #      delta:正則化項のパラメータ，
  #      tol:ブロック座標降下法の停止基準)
  # 出力(f:線形判別関数, lab:ラベル, worst:最悪の予測誤差の推定値)
  n <- nrow(x); d <- ncol(x)
  x1 <- x[y==levels(y)[1],]; x2 <- x[y==levels(y)[2],]
  m1 <- colMeans(x1); v1 <- cov(x1)
  m2 <- colMeans(x2); v2 <- cov(x2)
  a0 <- (m1-m2)/sum((m1-m2)^2)
  F0 <- diag(d)-(a0 %o% a0)/sum(a0^2)
  eigenF <- eigen(F0)
  F <- matrix(eigenF$vector[,order(eigenF$value)[-1]],d,d-1)
  G <- t(F) %*% v1 %*% F; H <- t(F) %*% v2 %*% F
  g <- t(F) %*% v1 %*% a0; h <- t(F) %*% v2 %*% a0
  # 初期化
  beta <- 1; eta <- 1; k <- 0
  err <- Inf; obj_pre <- Inf
  # ブロック座標降下法
  while((err > tol) & (k < maxit)){
    MLS <- G/beta+H/eta+delta*diag(d-1) # 3項目は正則化項
    bLS <- -(g/beta+h/eta)
    uLS <- solve(MLS,bLS)
    w <- a0 + F %*% uLS
    beta <- sqrt(sum((v1 %*% w) * w))
    eta <- sqrt(sum((v2 %*% w) * w))
    obj_val <- beta + eta + sum(w^2)*delta/2
    err <- abs((obj_pre - obj_val)/obj_val)
    obj_pre <- obj_val
    k <- k+1
  }
  # 推定値の計算
  b <- sum(w * m1) - beta/(beta+eta)
  gamma <- beta+eta
  worst_err <- (gamma^2)/(1+gamma^2)
  res <- list(f=function(x)x%*%w-b, lab=levels(y), worst=worst_err)
  class(res) <- 'mpm'
  res
}
predict.mpm <- function(res,new){
  # mpm でラベルを予測する関数
  # 入力:(res:関数 mpm の出力, new[行列]: 入力データ)
  # 出力:newに対する予測ラベル
  ypred <- factor(rep(NA,nrow(new)),levels=res$lab)
  yp <- res$f(new)
  ypred[yp>=0] <- res$lab[1]; ypred[yp<0] <- res$lab[2]
  ypred
}
